import copy
import datetime
import time

import numpy as np
import pandas as pd

from crossnetwork_processing_functions import get_data_files, create_GHOST_stage_logfile
from GHOST_standards import get_standard_data, get_standard_metadata, country_ISO3166_to_name, standard_data_flag_name_to_data_flag_code, standard_QA_name_to_QA_code
from network_file_standardise import NetworkFileStandardise
from standardisation_mapping_functions import map_network_classifications_to_standard_classifications, map_network_measurement_information_to_standard_measurement_information, network_data_flag_to_standard_data_flag_mapping
from time_functions import round_to_minute, datetime_to_fractional_minutes_from_reference, get_datafile_download_timestamp
import unit_converter

class StandardiseNetworkData:

    '''Convert 1 provided data file to GHOST standard numpy station specific files'''

    def __init__(self, null, run_details, parameter_details, network_details):
        
        #add passed variables to self
        self.run_details = run_details
        self.parameter_details = parameter_details
        self.network_details = network_details

    def do_work(self):
        '''Function which does executes the functions of the class in a pipeline'''

        #get data file to parse
        data_file = self.gather_data_files()

        #parse station reference data files
        for data_file in data_files:
            self.parse_file(data_file)

        #if stage has completed successfully, with some valid data, write a valid process log file
        if self.written_data:
            self.run_details['log_file_str'] += str((time.time() - self.run_details['stage_process_start_time'])/60.)
            create_GHOST_stage_logfile(0,self.run_details)
        #else, if no data has been written for station write an error log file to inform station is not continuing in the pipeline
        elif not self.written_data:
            self.run_details['log_file_str'] += 'NO DATA HAS BEEN WRITTEN FOR STATION'
            create_GHOST_stage_logfile(1,self.run_details)

    def gather_data_files(self):

        '''Function to gather data files to parse per station'''

        #get data files specific to station
        data_file = get_data_files(self.run_details, self.parameter_details)[0]   

        #load external metadata into memory 
        self.external_metadata = np.load(self.run_details['data_root']+'US_NADP_AMoN/metadata/processed/US_NADP_AMoN_META.npy', allow_pickle=True)

        #get standard measured unit quantity
        conv_obj = unit_converter.convert_units(self.parameter_details['standard_units'],self.parameter_details['standard_units'],1)
        self.measured_unit_quantity = conv_obj.output_represented_quantity

        return data_files

    def parse_file(self, data_file):

        '''Function to parse data file to GHOST standard numpy station specific file'''

        self.run_details['log_file_str'] += 'PROCESSING FILE: {}\n'.format(data_file) 

        #convert metadata data with units to standard units
        
        #get measurement units
        units = site_nasa_ames_obj.primary_units[var_ind].strip()
            
        #convert raw units given to standard name
        try:
            conv_obj = unit_converter.convert_units(units,units,1, measured_species = self.parameter_details['chemical_formula'])
        #if cannot convert units to standard units, then then a chemical formula is needed for conversion of units but do not have it: ug C/m3 for NMHC
        #therefore do not process file
        except SystemExit:
            self.run_details['log_file_str'] += 'NO VALID UNITS IN FILE: {}. SKIPPING.\n'.format(data_file)
            return
        site_metadata['measuring_instrument_reported_units']['value'][-1] = conv_obj.output_standard_units
        measuring_instrument_reported_unit_quantity = conv_obj.output_represented_quantity      

        #----------------------------------------------------------------------------#
        #READ FILE
        #----------------------------------------------------------------------------#

        site_nasa_ames_obj = ParseNASAAmesFile(data_file)
        site_nasa_ames_obj.read_data()

        #get all unique station references
        unique_station_references = np.unique(site_nasa_ames_obj.parsed_data['SITEID']])
        
        #define raw measurement units, and convert to standard name
        raw_units = 'ug/m3'
        try:
            conv_obj = unit_converter.convert_units(units,units,1, measured_species = self.parameter_details['chemical_formula'])
        #if cannot convert units to standard units, then then a chemical formula is needed for conversion of units but do not have it: ug C/m3 for NMHC
        #therefore do not process file
        except SystemExit:
            self.run_details['log_file_str'] += 'NO VALID UNITS IN FILE: {}. SKIPPING.\n'.format(data_file)
            return
        #site_metadata['measuring_instrument_reported_units']['value'][-1] = conv_obj.output_standard_units
        #measuring_instrument_reported_unit_quantity = conv_obj.output_represented_quantity      

        #iterate through unique station references
        for unique_station_reference in unique_station_references:

            start_timer = time.time()

            #set flag to know when have written data for station 
            written_data = False

            self.run_details['log_file_str'] += '{} -- PARSING RAW DATA\n'.format(station_reference)

            #create site instances of metadata/data dictionaries
            site_data =     get_standard_data(self.parameter_details)
            site_metadata = get_standard_metadata(self.parameter_details)

            #isolate variable data specific for station
            unique_station_bool = site_nasa_ames_obj.parsed_data['SITEID'] == unique_station_reference
            station_var_data = {var:site_nasa_ames_obj.parsed_data[var][unique_station_bool] for var in site_nasa_ames_obj.parsed_data.keys()}

            #------------------------------------------------------------------------#
            #PARSE TIME
            #------------------------------------------------------------------------#
        
            #parse start, end and modification datetime for each reported timepoint 
            #first create a unified date/time string (yyyy-mm-dd hh:mm:ss Z, of the measurement --> z is always GMT)
            #secondly parse into datetime (taking into account timezone information). then convert to UTC datetime, then round to nearest minute    
            start_dt = [round_to_minute(dateutil.parser.parse('{}:00 GMT'.format(date), tzinfos=tzinfos).astimezone(dateutil.tz.tzutc())).replace(tzinfo=None) for date in station_var_data['STARTDATE']] 
            end_dt = [round_to_minute(dateutil.parser.parse('{}:00 GMT'.format(date), tzinfos=tzinfos).astimezone(dateutil.tz.tzutc())).replace(tzinfo=None) for date in station_var_data['ENDDATE']] 
            mod_dt = [round_to_minute(dateutil.parser.parse('{}:00 GMT'.format(date), tzinfos=tzinfos).astimezone(dateutil.tz.tzutc())).replace(tzinfo=None) for date in station_var_data['CHNGDATE']] 
                        
            #convert each start measurement datetime to minutes from reference time
            frac_time = [datetime_to_fractional_minutes_from_reference(dt) for dt in start_dt]

            #get timedeltas between start and end of measurement window
            temporal_resolutions = []
            for start, end in zip(start_dt,end_dt):
                dt_window = end - start
                #convert hours to minutes
                dt_window_minutes = int(dt_window.total_seconds()/60.)
                temporal_resolutions.append(dt_window_minutes)  

            #------------------------------------------------------------------------#
            #PARSE QA/FLAGS
            #------------------------------------------------------------------------#

            flags = []
            qa = []

            #get dictionary necessary for mapping of US NADP AMoN file data flags to standard data flag codes
            data_flag_mapping_dictionary = network_data_flag_to_standard_data_flag_mapping['US_NADP_AMoN']                

            #get network data flags from data file
            network_data_flags = station_var_data['QR']

            #get network notes associated informing on each data flag
            network_notes = station_var_data['NOTES']

            #iterate through network data flags and network notes
            for network_data_flag, network_note in zip(network_data_flags, network_notes):

                #lists to append qa/data flags per measurement
                flags_per_measurement = []
                qa_per_measurement = []

                #set GHOST decreed validity/network decreed validity variables initially as 'V'
                ghost_decreed_validity = 'V'
                network_decreed_validity = 'V'

                #iterate through all network note codes:
                for network_note_code in network_note:
                    
                    #concatenate network_data_flag + network_note_code
                    data_flag = network_data_flag + network_note_code

                    # map data flag/s and append
                    standard_data_flag_dict = data_flag_mapping_dictionary[data_flag]
                    #get mapped standard data flags
                    standard_data_flag_names = standard_data_flag_dict['standard_data_flag_name']
                    #iterate through standard data flags 
                    for standard_data_flag_ii, standard_data_flag_name in enumerate(standard_data_flag_names):
                        #check if standard data flag is empty (i.e. unmapped as non-useful)
                        if standard_data_flag_name != '': 
                            #get corresponding data flag code for standard data flag
                            flag_code = standard_data_flag_name_to_data_flag_code[standard_data_flag_name]
                            #append standard data flag
                            flags_per_measurement.append(flag_code)
                            #determine if GHOST/network decreed valid
                            if standard_data_flag_dict['GHOST_decreed_validity'][standard_data_flag_ii] == 'I':
                                ghost_decreed_validity = 'I'
                            if standard_data_flag_dict['network_decreed_validity'][standard_data_flag_ii] == 'I':
                                network_decreed_validity = 'I'
    
                #if ghost decreed validity, or network decreed validity is Invalid, then set QA flags
                if ghost_decreed_validity == 'I':
                    qa_per_measurement.append(standard_QA_name_to_QA_code['Invalid Data Provider Flags - GHOST Decreed'])
                if network_decreed_validity == 'I':
                    qa_per_measurement.append(standard_QA_name_to_QA_code['Invalid Data Provider Flags - Network Decreed'])
        
                #append qa_per_measurement, flags_per_measurement to lists to write out
                flags.append(flags_per_measurement)
                qa.append(qa_per_measurement)

            #------------------------------------------------------------------------#
            #PARSE METADATA
            #------------------------------------------------------------------------#

            #GHOST version
            site_metadata['GHOST_version']['value'].append(self.run_details['ghost_version'])

            #initialise string which will store all GHOST process warnings as empty            
            site_metadata['process_warnings']['value'].append('')

            #check if measurement position metadata is suspect and write it to metadata/data dictionaries
            site_metadata['longitude']['value'].append()
            site_metadata['latitude']['value'].append()
            site_metadata['altitude']['value'].append()

            #save meta_update_stamp, set as copy of time array
            site_metadata['meta_update_stamp']['value'] = sorted(frac_time)[0]

            #write known metadata  
            site_metadata['network']['value'].append(self.run_details['network'])
            site_metadata['station_name']['value'].append()
            site_metadata['data_level']['value'].append('1.0')
            site_metadata['contact_name']['value'].append() 
            site_metadata['contact_institution']['value'].append()
            site_metadata['contact_email_address']['value'].append()
            site_metadata['principal_investigator_name']['value'].append()
            site_metadata['principal_investigator_institution']['value'].append()
            site_metadata['principal_investigator_email_address']['value'].append()
            #assume datum information
            site_metadata['ellipsoid']['value'].append('WGS 84')
            site_metadata['horizontal_datum']['value'].append('World Geodetic System 1984')
            site_metadata['vertical_datum']['value'].append('tidal - mean sea level') 
            site_metadata['projection']['value'].append('geographic')

            #process measurement information
            standard_measurement_metadata, measurement_qa_flags, standard_method_abbreviation = map_network_measurement_information_to_standard_measurement_information(self.parameter_details['aeronet_matrix'], site_metadata, self.run_details, self.parameter_details, write_to_metadata_array=False)
            for key, value in standard_measurement_metadata.items():
                site_metadata[key]['value'] = [value] * len(frac_time)
            site_metadata['measurement_qa_flags'] = {'value':[], 'data_type':np.object, 'string_format':'lower_short'}
            measurement_qa_flag_codes = []
            #set measurement qa codes which are True as joined metadata str
            for standard_flag_name in measurement_qa_flags.keys():
                if measurement_qa_flags[standard_flag_name]:
                    measurement_qa_flag_codes.append(str(standard_QA_name_to_QA_code[standard_flag_name]))
            site_metadata['measurement_qa_flags']['value'] = [','.join(measurement_qa_flag_codes)]*len(frac_time)

            #append station_reference now that we have the standard method abbreviation
            site_metadata['station_reference']['value'] = ['{}_{}'.format(station_reference, standard_method_abbreviation)]*len(frac_time)

            #iterate through metadata variables, if any columns have not been appended to add a NaN
            for column_header in list(site_metadata.keys()):
                n_column = len(site_metadata[column_header]['value'])
                #if column is empty, then append NaN
                if n_column < 1:
                    site_metadata[column_header]['value'] = [np.NaN] * len(frac_time)

            #get data file download timestamp
            site_metadata['data_download_stamp']['value'] = [get_datafile_download_timestamp(data_file)] * len(frac_time)

                

